// This  skeleton file was automatically created by 
//            pkgmaker 0.1
//            Thursday, December 11 1997
// 
// Module: ConcreteArfDataServer
// Author: Hassan Siddiqui
//
// Description: Concrete implementation of ArfDataServer class
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: ConcreteArfDataServer.cc,v 1.33 2000/12/19 14:00:15 hsiddiqu Exp $

// General classes
#include <cmath>

// SAS-specific classes
#include "CalServer.h"
#include "Dal.h"
#include "OalServer.h"
#include "param.h"
#include "Region.h"
#include <ASCregion.h>
#include "ExceptionErrorHandler.h"
#include "DataFilter.h"
#include "dss.h"
#include "DssBuilder.h"

// General utility functions
#include "Tracer.h"

// Local Classes
#include "ConcreteArfDataServer.h"
#include "PsfModel.h"
#include "CalPsfModel.h"
#include "RatioPsfModel.h"
#include "CalMOSCoreData.h"
#include "CalPNCoreData.h"
#include "CheckExpr.h"
#include "GeneralIterator.h"
#include "TablePhotonDistribution.h"
#include "NullPhotonDistribution.h"
#include "Frame.h"
#include "ArrayPixelIterator.h"
#include "DummyPixelIterator.h"
#include "NullPixelIterator.h"
#include "PsfmodelPixelIterator.h"
#include "NullEeCorrector.h"
#include "PointEeCorrector.h"
#include "RmfEbins.h"
#include "RegionExtent.h"
#include "BadPixelCorrector.h"
#include "NullBadPixelCorrector.h"
#include "OffCCDPixelIterator.h"

/* Constructor */


ConcreteArfDataServer::ConcreteArfDataServer(): 
  _rawDetmap(0),
  _coreData(0),
  _photonDistribution(0),
  _extendedsource(booleanParameter("extendedsource")),
  _modelEffectiveArea(booleanParameter("modeleffarea")),
  _modelQuantumEfficiency(booleanParameter("modelquantumeff")),
  _modelFilterTransmission(booleanParameter("modelfiltertrans")),
  _modelootcorr(booleanParameter("modelootcorr")),
  _useodf(booleanParameter("useodfatt")),
  _eeCorrector(0),
  _extent(0)
{
  PROC;

  /* Filtered table settings */
  _filterdss = booleanParameter("filterdss");
  _keepfilteredtable=booleanParameter("withfilteredset");
  _filteredtablename=stringParameter("filteredset");

      
  /* Open spectrum dataset - access=read (modify if BACKSCAL to be written) */
  string spectrumsetname = stringParameter("spectrumset");
  errstr 
    << "Opening spectrumset " << spectrumsetname << "..."
    << Sas::message(Msg::App, Msg::Sparse);
  
  DataSet * spectrumset;

  if (booleanParameter("setbackscale")) {
     spectrumset = dataSetServer->open( spectrumsetname, 
			 DataSet::Modify );
  } else {
     spectrumset = dataSetServer->open( spectrumsetname,
                         DataSet::Read );
  }
  
  /* Access spectrum  */
  Table * spectrumtable = spectrumset->table("SPECTRUM"); 
  
  /* Get observation start time */
  _dateobs = spectrumset->attribute("DATE-OBS")->asString();

  /* Get pointing info */
  _rapnt = spectrumset->
    attribute("RA_PNT")->
    asReal();

  _decpnt = spectrumset->
    attribute("DEC_PNT")->
    asReal();

  _papnt = spectrumset->
    attribute("PA_PNT")->
    asReal();

  /* Get WCS information for X/Y columns */
  _poswcs = new PosWCS(spectrumset, _rapnt, _decpnt, _papnt, _useodf);
  _refcoords = new CoordPair(_poswcs->wcs()->physicalRefCoords());

  /* Get telescope */
  _telescope = spectrumset->
    attribute("TELESCOP")->
    asString();

  /* Get datamode */
  _datamode = spectrumset->
    attribute("DATAMODE")->
    asString();
    
  /* Get observingmode */
  if (spectrumset->hasAttribute("SUBMODE")) 
  {
     _windowMode = getWindowMode( spectrumset->attribute("SUBMODE")->
                                                   asString() );
errstr<<"Window mode: "<<_windowMode
<< Sas::message(Msg::App, Msg::Verbose);

  } else {
     _windowMode = getWindowMode(" ");
  }
    
  /* Get filter */
  _filter = spectrumset->attribute("FILTER")->asString();

  /* If the filter is UNKNOWN then exit 
     unless this is a backscale run */
  if (_filter == "UNKNOWN")
  {
    if (! booleanParameter("keeparfset")) {
      _filter="Medium";
    } else {
      errstr<<"Filter not set in file header"
            << Sas::error("BadFilterName");
    }
  }

  /* Get Pipeline version */
  if (spectrumset->hasAttribute("PPSVERS"))
  {
     _ppsvers = spectrumset->
        attribute("PPSVERS")->
        asString();
  } else {
     _ppsvers = "NotSet";
  }

  /* If the filter is bad the code exits later with a message to 
     this effect. If this is a backscale run (i.e. the ARF wont be kept 
     - set the filter to medium just in case */
  if (! booleanParameter("keeparfset")) {
      _filter="Medium";
  }

  /* Get selection expression */
  if (spectrumtable->hasAttribute("SLCTEXPR"))
    {
      _slctexpr = spectrumtable->
	attribute("SLCTEXPR")->
	asString();
    }
  else
    {
      errstr 
	<< "Could not find an attribute SLCTEXPR in table "
	<< spectrumtable->qualifiedName()
	<< ", so assuming SLCEXPR to be true "
	<< Sas::message(Msg::App, Msg::Noisy);

      _slctexpr = "true";
    }

  /* Get also DSS info */
  _dss = new DataSubspace(spectrumtable);

errstr<<"DSS: components in Concrete: "<<_dss->components()
<< Sas::message(Msg::App, Msg::Verbose);

  /* Check that these is some dss info present */
  if (_dss->components() == 0) {
      errstr
        << "No datasubspace info found in spectrum"
        << Sas::error("DSSNotFound");
  }
      
  /* Read useful quantities from the DSS */
  readDSS();

  /* Get instrument ID */
  _instrument = spectrumtable->
    attribute("INSTRUME")->
    asString();
  
  _instrumentID =  Xmm::toInstrument(_instrument);
  
  errstr << "Got instrument "
	 << _instrument
	 << " which has ID "
	 << _instrumentID
	 << Sas::message(Msg::App, Msg::Noisy);

  /* Spectrum table no longer needed. Close it */
  errstr 
    << "Closing spectrumset " << spectrumsetname << "."
    << Sas::message(Msg::App, Msg::Sparse);
  dataSetServer->close(spectrumset);

  double psfenergy = realParameter("psfenergy") * 1000; // in eV.

  /* Set some CAL quantities which apply to both instruments */
  calServer.setInstrument(_instrumentID);
  calServer.state()->filter()->set(Xmm::toFilter(_filter));
  calServer.state()->mode()->set(_windowMode);
  calServer.state()->randomize()->set(false); // Switch off randomization

  /* Now set the instrument specific CAL quantities */
  switch (_instrumentID)
    {
    case Xmm::EMOS1 :
    case Xmm::EMOS2 :
      {
	Sas::message(Msg::App,
		     Msg::Verbose,
		     "Creating a MOSCoreData object");
	_coreData = 
	  new CalMOSCoreData( _instrumentID,
			      _datamode,
			      _dateobs,
			      _poswcs,
			      _filter,
			      psfenergy); 
	break;
      }
    case Xmm::EPN :
      {
	Sas::message(Msg::App,
		     Msg::Verbose,
		     "Creating a PNCoreData object");
	_coreData = 
	  new CalPNCoreData( _instrumentID,
			     _datamode,
			     _dateobs,
			     _poswcs,
			     _filter,
			     psfenergy); 
	break;
      }
    default:
      errstr << "can't generate arf data for instrument of type " 
	     << _instrument 
	     << Sas::error("invalidInstrument");
    }

  /* For timing and burst modes the spatial selection is either left blank 
     or is made in RAW-X. In this case a pseudo-region should be generated
     in DET coords to handle vignetting and PSF issues */

  _TimOrBurstMode=false;
  if (_windowMode == Xmm::FAST_TIMING || _windowMode == Xmm::FAST_BURST
                                     || _windowMode == Xmm::FAST_UNCOMPRESSED 
                                     || _windowMode == Xmm::FAST_COMPRESSED) {
     _TimOrBurstMode=true;
     addDetRegion(_coreData->varpositionUtils());
  }

  /* Find from the DSS if it is a POS (X/Y) region - this means that no
      mixing of X/Y DETX/Y filters is possible which is poor */
  DSS::DSaxes axestest;
  axestest.push_back("POS");
  _regPosCoords = ! CheckExpr::check(_dss->selectExpression(axestest));

  errstr<<"regPos= "<<_regPosCoords
        << Sas::message(Msg::App, Msg::Verbose);

  /* Set the extent of the source box in detector coords */
  _extent = new RegionExtent(_dss, _coreData->varpositionUtils(),
               _datamode, _dateobs, _poswcs);

  /* Create source position object */
  bool withsourcepos = booleanParameter("withsourcepos");
  string sourceCoordType=" ";
  double sourcex=0.0;
  double sourcey=0.0;

  /* [HS] START Modifed withsourcepos logic */
  if (withsourcepos)
    {
      sourceCoordType = uppercase( stringParameter("sourcecoords") );
      sourcex = realParameter("sourcex");
      sourcey = realParameter("sourcey");
    }
  else
    {
      /* If the source position hasn't been specified,
         set it in detector
         coords as the centre of the source region */

      errstr
        << "Computing source position from region extent.."
        << Sas::message(Msg::App, Msg::Sparse);

      sourceCoordType = "DET";

      if (_extent->isValid())
        {
          sourcex = (_extent->xmin() + _extent->xmax()) / 2.0;
          sourcey = (_extent->ymin() + _extent->ymax()) / 2.0;
        }
      else
        {
          errstr
             << "Cannot determine source position from DSS entry"
             << Sas::error("RegionInvalid");
        }
    }

  /* Set source position */
  _coreData->setSourcePosition(sourceCoordType, sourcex, sourcey);

  /* Inform user of source position */
  errstr
    << "Using a source position in " << sourceCoordType
    << " coordinates of x=" << sourcex
    << " y=" << sourcey
    << Sas::message(Msg::App, Msg::Sparse);

  if (sourceCoordType != "DET")
    {
      errstr
        << "This corresponds to a position in DET coordinates of "
        << "DETX=" << _coreData->sourcePosition()->xdet()
        << " DETY=" << _coreData->sourcePosition()->ydet()
        << Sas::message(Msg::App, Msg::Sparse);
    }
 
  errstr
    << " The source position in focal plane coords is "
    << " theta=" << _coreData->sourcePosition()->theta()
    << " phi=" << _coreData->sourcePosition()->phi()
    << Sas::message(Msg::App, Msg::Sparse);

  /* [HS] END withsourcepos logic */

  /* Initialise pattern list using selection in DSS. 
     This could have been done via
     CoreData object constructor, but this approach minimises
     messy parameter propagation. */
  
  _coreData->initPatterns(_dss);
      
  /* Set-up the point spread function */
  _coreData->setPSF(region(), _extent);

  /* Create the detector map */
  createDetmap();
	    
  /* Create DetmapFilter object. 
     This will be used once the ArfDataServer is fully constructed. */
  _detmapfilter = new DetmapFilter;
  _filteredDetmapExists = false;

  /* Create PhotonDistribution object */
//    if (booleanParameter("modelpileup"))
//      { }
//    else
//      {
      _photonDistribution = 
	new NullPhotonDistribution();
//      }
  
  /* Create frame object */
  _frame = new Frame( _photonDistribution );

  _sourceBoxArea = (new OffCCDPixelIterator(
                   regionToPoly(region()),
                   _extent,
                   positionUtils(),
                   ignoreEdgePixels() ))->totalPixels();

  errstr<<"Source box size: "<<_sourceBoxArea
  << Sas::message(Msg::App, Msg::Verbose);

  /* Create a bad pixel corrector if required */
  if (booleanParameter("withbadpixcorr"))
    {

      /* Create a correction vector for the bad pixels depending on
         whether the source is deemed to be point-like or extended */

      if (_extendedsource) {

         _bp_psf = new RatioPsfModel( _sourceBoxArea );

      } else { // (point source: so use the same PSF as for modelee corrs.)

         _bp_psf = _coreData->psf();
      }

      _badPixCorrector = new BadPixelCorrector(
                            stringParameter("badpixlocation"),
                            (unsigned) intParameter("eegridfactor"), _bp_psf);
    }
  else {
      _badPixCorrector = new NullBadPixelCorrector();
  }

  /* Create Ebins object, depending on requested source */
  if (booleanParameter("withrmfset"))
    {
      _ebins = RmfEbins::ebins(stringParameter("rmfset"));
    }
  else
    {
      _ebins = _coreData->ebins();
    }

  /* Create EeCorrector object, if source isn't extended and this isn't
     Timing or Burst mode */
  if (booleanParameter("modelee") 
      && (!_extendedsource) 
      && (!_TimOrBurstMode) )
    {
      _eeCorrector = new PointEeCorrector( (unsigned) intParameter("eegridfactor") );

     /* Tell user which PSF is being used */
      errstr << "PSF type: "<<calServer.state()->accuracyLevel()->get()<<" will be used"
             << Sas::message(Msg::App, Msg::Sparse);
    }
  else
    {
      _eeCorrector = new NullEeCorrector();
    }

  /* Set up a constant curve objects 
     in case relevant model is switched off */
  
  vector<double> egrid;
  egrid.push_back(0.0);
  egrid.push_back(40000.0);
  
  vector<double> eeff;
  eeff.push_back(1.0);
  eeff.push_back(1.0);
  
  _nulleacurve = new EffectiveAreaCurve(egrid, eeff);
  _nullqecurve = new CcdQuantumEfficiencyCurve(egrid, eeff);
  _nullftcurve = new FilterTransmissionCurve(egrid, eeff);

  ENDPROC;
};
  
/* Destructor */
ConcreteArfDataServer:: ~ConcreteArfDataServer()
{
  PROC;
  
  delete _nullftcurve;          _nullftcurve          = 0;
  delete _nullqecurve;          _nullqecurve          = 0;
  delete _eeCorrector;          _eeCorrector          = 0;
  delete _ebins;                _ebins                = 0;
  delete _frame;                _frame                = 0;
  delete _photonDistribution;   _photonDistribution   = 0;
  delete _detmapfilter;         _detmapfilter         = 0;
  delete _rawDetmap;            _rawDetmap            = 0;
  delete _coreData;             _coreData             = 0;
  delete _dss;                  _dss                  = 0;
  delete _poswcs;               _poswcs               = 0;
  delete _extent;               _extent               = 0;

  ENDPROC;
};

  /********************************************
   *
   * General observation/mission/instrument info
   *
   ********************************************/

Xmm::Instrument ConcreteArfDataServer::instrumentID()  const
{
  PROC;
  RETURN _instrumentID;
}

const string   ConcreteArfDataServer::instrument()   const
{
  PROC;
  RETURN _instrument;
}

const string   ConcreteArfDataServer::telescope()   const
{
  PROC;
  RETURN _telescope;
}

const string   ConcreteArfDataServer::datamode()   const
{
  PROC;
  RETURN _datamode;
}

const string   ConcreteArfDataServer::dateobs()   const
{
  PROC;
  RETURN _dateobs;
}

const string   ConcreteArfDataServer::filter ()    const
{
  PROC;
  RETURN _filter;
}


/**************************
 *
 * Pointing info
 *
 **************************/
  
/* Source position */
double  ConcreteArfDataServer::raref()  const
{
  PROC;
  RETURN _refcoords->x();
}

double  ConcreteArfDataServer::decref() const
{
  PROC;
  RETURN _refcoords->y();
}

/* Mean s/c pointing */
double  ConcreteArfDataServer::rapnt()  const
{
  PROC;
  RETURN _rapnt;
}

double  ConcreteArfDataServer::decpnt() const
{
  PROC;
  RETURN _decpnt;
}

double  ConcreteArfDataServer::papnt()  const
{
  PROC;
  RETURN _papnt;
}

bool ConcreteArfDataServer::useODF() const
{
  PROC;
  RETURN _useodf;
}

/* Method to say if pixels on the edges of CCDs are excluded in the spectrum */
bool ConcreteArfDataServer::ignoreEdgePixels() const
{
  PROC;

 /* The default parameters are that MOS edge pixels are included in the
    spectrum and PN pixels are excluded. This can be changed if non-default
    flags are used, e.g. flag=0 - handle this later */

  bool ignore=false;

 /*
    switch (_instrumentID)
      {
      case Xmm::EMOS1 :
      case Xmm::EMOS2 :
        {
          ignore=false;
	  break;
        }
      case Xmm::EPN :
        {
          ignore=true;
	  break;
        }
      default:
          errstr << "Instrument unknown"
	     << _instrument 
	     << Sas::error("invalidInstrument");
      }
  */

  /* Check flag value - if zero then edge pixels are ignored. We should
     really check for the actual bit value but I can't find that info. */
      if (_flag == 0) {
         ignore=true;
      }

      errstr<<"Ignore flag set to: "<<ignore
            <<" Flag= "<<_flag
        << Sas::message(Msg::App, Msg::Verbose);

  RETURN ignore;
}

/********************************************
 *
 * Selection info
 *
 ********************************************/

const string &   ConcreteArfDataServer::selectionExpression()  const
{
  PROC;
  RETURN _slctexpr;
}

const DataSubspace * const  ConcreteArfDataServer::dataSubspace()  const
{
  PROC;
  RETURN _dss;
}

const RegionExtent * ConcreteArfDataServer::extent()  const
{
  PROC;
  RETURN _extent;
}

/* Method to convert XY position to DETX,DETY */
SourcePosition ConcreteArfDataServer::xyToDetxy (double sourcex,
                                                double sourcey) const
{
  PROC;

     errstr<<"Entered xyToDetxy: "<<sourcex<<" "<<sourcey
    << Sas::message(Msg::App, Msg::Verbose);

 /* This uses a temporary positionUtils object. This must change and
    is pending a restructuring */

  PositionUtils * testpos = new PositionUtils(_instrumentID);

  SourcePosition s = testpos->sourcePosition( "POS",
                                                     _datamode,
                                                     _dateobs,
                                                     sourcex,
                                                     sourcey,
                                                     _poswcs);
  RETURN s;
}


/********************************************
 *
 * Ebins accessor
 *
 ********************************************/

const BinnedAxis * const    ConcreteArfDataServer::ebins()   const
{
  PROC;
  RETURN _ebins;
}

/********************************************
 * 
 * Source Position
 *
 ********************************************/

const SourcePosition * const  ConcreteArfDataServer::sourcePosition() const
{
  PROC;
  RETURN _coreData->sourcePosition();
}


/********************************************
 *
 * Misc
 *
 ********************************************/

bool ConcreteArfDataServer::filterDSS ()    const
{
  PROC;
  RETURN _filterdss;
}


const string ConcreteArfDataServer::filteredTableName ()    const
{
  PROC;
  RETURN _filteredtablename;
}

bool ConcreteArfDataServer::keepFilteredTable ()    const
{
  PROC;
  RETURN _keepfilteredtable;
}

bool ConcreteArfDataServer::isExtendedSource() const
{ PROC; RETURN _extendedsource;}

const PhotonDistribution * const  
ConcreteArfDataServer::photonDistibution() const
{
  PROC;
  RETURN _photonDistribution;
}

const Frame * const  ConcreteArfDataServer::frame() const
{
  PROC;
  RETURN _frame;
}
  
PsfModel * const  ConcreteArfDataServer::psf() const
{
  PROC;
  RETURN _coreData->psf();
}

PixelIterator * const  ConcreteArfDataServer::detmapPixelIterator()
{
  
  PROC;

  /* Ideally, the filtered detmap would be created
     during construction of the arfDataServer,
     but the process unfortunately needs
     other quantities from the server. */

    if (! _filteredDetmapExists) {
       _filteredDetmap = _detmapfilter->filteredMap(_rawDetmap);
       _filteredDetmapExists = true;
    }

  RETURN _filteredDetmap;
}

/* Provide a pixel iterator to evaluate the PSF function*/
PixelIterator * const  ConcreteArfDataServer::PsfPixelIterator()
{

  PROC;

  /* If this is a circular PSF function then we just need a dummy one
     pixel iterator. Otherwise integrate over the detector map */

     errstr<<"Acclevel is: "<<calServer.state()->accuracyLevel()->get()
    << Sas::message(Msg::App, Msg::Verbose);

    if ( calServer.state()->accuracyLevel()->get() == EXTENDED)
    {
     errstr<<"Acclevel dum"
    << Sas::message(Msg::App, Msg::Verbose);
       RETURN (new DummyPixelIterator( 1.0, 2.0, 1, 1.0, 2.0, 1 ));
    }
    else {
     errstr<<"Acclevel detmap"
    << Sas::message(Msg::App, Msg::Verbose);
       RETURN detmapPixelIterator();
    }
}

double ConcreteArfDataServer::framePeriod()    const
{
  PROC;
  RETURN _coreData->framePeriod();
}

bool ConcreteArfDataServer::setCcd( const CamCoords & camcoord)
{
  PROC;
  RETURN _coreData->setCcd(camcoord);
}


/* Convert from CAMCOORD2 to PIXCOORD */
const PixCoords ConcreteArfDataServer::camCoord2ToPixCoord( const CamCoords & camcoords ) const
{
      PROC;
      RETURN _coreData->camCoord2ToPixCoord(camcoords);
}

/* Return the positionUtils object */
const PositionUtils * ConcreteArfDataServer::positionUtils() const
{ PROC;
   RETURN _coreData->positionUtils();
}

/* Convert from DETX/DETY (0.05 arcsec units) to camcoord2 */
const CamCoords ConcreteArfDataServer::detxy005ToCamCoord2(int detx_005, 
							   int dety_005) const
{
  PROC;
  RETURN _coreData->detxy005ToCamCoord2( detx_005,
					 dety_005 );
}

/* Convert from CAMCOORD2 to TELCOORD */
const TelCoords ConcreteArfDataServer::camCoord2ToTelCoord( const CamCoords & camcoords ) const
{
  PROC;
  RETURN _coreData->camCoord2ToTelCoord(camcoords);
}


  
/*********************** 
 *  Arf quantity routines 
 ***********************/

/* These methods can be turned on/off by user */
EffectiveAreaCurve ConcreteArfDataServer::effectiveAreaCurve ( double theta, 
                double phi ) const
{
  PROC;
  if (!_modelEffectiveArea) ALTRETURNV( *_nulleacurve );
  
  //double area;
  EffectiveAreaCurve eacurve;

/* For point sources return the effective area at the source position */
  if (! _extendedsource) {
     eacurve=_coreData->effectiveArea();
     errstr<<"Calculating eff. area at source pos"
    << Sas::message(Msg::App, Msg::Verbose);
  } 
/* For extended sources return effective area at the input position */
  else {
     eacurve=_coreData->effectiveArea(theta, phi);
  }

  RETURN eacurve;
}

CcdQuantumEfficiencyCurve
ConcreteArfDataServer::quantumEfficiencyCurve   ( double  x , 
						  double  y  ) const
{
  PROC;
  if (!_modelQuantumEfficiency)
    {
      ALTRETURNV( *_nullqecurve );
    }

  errstr 
    << "Getting QE curve for position " 
    << x
    << ","
    << y
    << "..."
    << Sas::message(Msg::App, Msg::Noisy);

  CcdQuantumEfficiencyCurve qecurve = 
    _coreData->quantumEfficiencyCurve(x,y);

  errstr 
    << "Got QE curve. " 
    << Sas::message(Msg::App, Msg::Noisy);
  
  RETURN qecurve;

}

FilterTransmissionCurve
ConcreteArfDataServer::filterTransmissionCurve   ( double  theta , 
						   double  phi  ) const
{
  PROC;
  if (!_modelFilterTransmission)
    {
      ALTRETURNV( *_nullftcurve );
    }

  FilterTransmissionCurve ftcurve = 
    _coreData->filterTransmissionCurve(theta, phi);

  RETURN ftcurve;

}


double ConcreteArfDataServer::pileupCorrection   (double  energy,
						  double  pixelFlux) const
{
  PROC;
  
  double corrn =  _coreData-> 	
    pileupCorrector()->
    correction(energy, pixelFlux);
  
  RETURN corrn;
}


/* Method to return the out-of-time correction value */
double ConcreteArfDataServer::ootCorrection ( ) const
{
  PROC;

  if (!_modelootcorr)
    {
      ALTRETURNV( 1.0 );
    }

  double ootfact=1.0;

/* Get the pipeline date from the ppsvers string */
  string ppsdate;

  if (_ppsvers != "NotSet") 
  {
     ppsdate = _ppsvers.substr(_ppsvers.find("/"),_ppsvers.find("."));
  }
  else {
     ppsdate = "99999999";

     errstr<<"No PPS version present in header "
             <<"not applying OOT correction"
	       << Sas::warning("NoPPSVersion");
  }

/*  The algorithm is that the out-of-time effect is zero unless this
    is a PN observation in either FF or EFF mode taken from a spectrum
    produced with an earlier pipeline version than 18-June-2001.
    MOS oot factors are included in the exposure time always. */

  if (_instrumentID == Xmm::EPN && ppsdate < "20010618" )
    {
      switch (_windowMode) {

        case Xmm::PRIME_FULL_WINDOW:
          {
            ootfact = 0.938;
            break;
          }
        case Xmm::PRIME_FULL_WINDOW_EXTENDED:
          {
            ootfact = 0.977;
            break;
          }
        case Xmm::PRIME_SMALL_WINDOW:
          {
            ootfact = 0.989;
            break;
          }
        default:
            ootfact = 1.0;
      }
    } else {
       ootfact = 1.0;
    }

   RETURN ootfact;
}


const EeCorrector * const ConcreteArfDataServer::eeCorrector () const
{
  PROC;
  RETURN _eeCorrector;
}

const BpCorrector * const ConcreteArfDataServer::badPixCorrector () const
{
  PROC;
  RETURN _badPixCorrector;
}

/* Method to find the observing window mode */
Xmm::Mode ConcreteArfDataServer::getWindowMode(string submode) const
{
  PROC;

      ExceptionErrorHandler local;
      ErrorHandler* oldHandler = errHandler.handler(&local);
      Xmm::Mode windowMode;
      try {
            windowMode = Xmm::toMode(submode);
      } catch(SasException::Error) { errstr<<"Submode not valid or missing - using FullWind
ow"
<<Sas::warning("UnknownModeString");
        windowMode = Xmm::PRIME_FULL_WINDOW; // d/f
      }
      errHandler.handler(oldHandler);

  RETURN windowMode;
}

/* Method to return the backscale value */
int ConcreteArfDataServer::backScale() const
{
  PROC;

   /* return the number of detector pixels in the source area minus
      the bad pixels and the chip gaps */ 
  int totarea = _sourceBoxArea - _badPixCorrector->totalPixels();
    errstr<<"RDS: box area "<<_sourceBoxArea
             <<"RDS: totpix "<< _badPixCorrector->totalPixels()
                  <<"RDS: totarea "<<totarea
	       << Sas::message(Msg::App, Msg::Verbose);

     
  RETURN totarea;
}

/* Method to return the DSS region as a region object */
GeoRegion 
ConcreteArfDataServer::region() const
{
  PROC;

    GeoRegion region;
    //GeoRegion tempregion;

   /* Get the filter corresponding to the region block within the dss */
    for (unsigned i  = 0; i < _dss->components(); ++i) {

      DScomp * dscomp = _dss->component(i);

     /* Loop over each filter in this component */
      for (unsigned f = 0; f < dscomp->filters(); ++f) {

          DSfilter * dsfilt = dscomp->filter(f);

          DSS::DSfilterComp caxes = dsfilt->comps();

        /* If the filter is a region - get a region object 
           Make sure that global filters are only included once.
           Exclude RAWX/Y regions */

           bool globalFilter = _dss->isGlobal((*dsfilt));
           if ( (! globalFilter || i==0) 
                      && (dsfilt->type() == DSS::RegionBlockT
                          || dsfilt->type() == DSS::RegionT) &&
                    caxes[0] != "RAWX" && caxes[1] != "RAWX"
                               ) {

              GeoRegion tempregion;

              if (dsfilt->type() == DSS::RegionBlockT) {

               /* Write region spec into a separate region object */
                InASCregionTable ascTab(dsfilt->blockSpec());
                ascTab >> tempregion;

              } else if (dsfilt->type() == DSS::RegionT) {
                tempregion = dsfilt->region();
              }

              for (unsigned comp=0; comp<tempregion.size(); comp++) {
                  region.push_back(tempregion[comp]);
              }

             errstr<<"region block added: " << dsfilt->selectExpression()
               <<" comp="<<i
	       << Sas::message(Msg::App, Msg::Verbose);
           }
      }
    }

    RETURN region;

}

/* Method to convert all circles to polygons within a region object */
GeoRegion 
ConcreteArfDataServer::regionToPoly(GeoRegion inRegion) const
{
  PROC;

   GeoRegion outRegion;

   for (unsigned comp=0; comp<inRegion.size(); comp++) {

      Component c; 

      for (unsigned element=0; element<inRegion[comp].size(); element++) {

        errstr<<"SHAPE is: "<<inRegion[comp][element]->type()
	       << Sas::message(Msg::App, Msg::Verbose);

      /* Copy element to output component, converting to polygons if necc. */

        vector<FShape *> newEls = shapeToPoly(inRegion[comp][element]);

        for (unsigned e=0; e<newEls.size(); e++) {
            c.push_back(newEls[e]);
        }

      }   // End of this component

     /* Push back and reset the component */
      outRegion.push_back(c);

      errstr<<"Added component: "<<comp
	       << Sas::message(Msg::App, Msg::Verbose);
   }

   RETURN outRegion;
}

/* Method: shapeToPoly(FShape element, Component c)
   Purpose: converts a DSS shape into polygons and adds to component
*/
vector<FShape *> ConcreteArfDataServer::shapeToPoly(FShape * element) const
{
  PROC;

 /* Pass the shape to the converter. If an annulus, convert into two 
    shapes */

    vector<FShape *> shapes;
    FShape * shape1;
    FShape * shape2;

    switch(element->type())
    {

     /* Convert annular shapes into two shapes */
      case ShapeBasics::ANNULUS :
      case ShapeBasics::ELLIPTANNULUS :
      {
         float cX = element->pointsX()[0];
         float cY = element->pointsY()[0];
         float a = element->radii()[0];
         float b = element->radii()[1];

         errstr<<"Ann centre: "<<cX<<" "<<cY
               <<"Radii: "<<a<<" "<<b
	       << Sas::message(Msg::App, Msg::Verbose);

         if (element->type() == ShapeBasics::ANNULUS) {
            shape1 = new FCircle(FPoint(cX, cY), a, ! element->included());
            shape2 = new FCircle(FPoint(cX, cY), b,  element->included());
         }
         else { // ELLIPTANNULUS:
            float c = element->radii()[2];
            float d = element->radii()[3];

            errstr<<"EllipAnn Radii: "<<c<<" "<<d
	       << Sas::message(Msg::App, Msg::Verbose);

            shape1 = new FEllipse(FPoint(cX, cY), a, b, ! element->included());
            shape2 = new FEllipse(FPoint(cX, cY), c, d, element->included());
         }

         shapes.push_back(shapeToPolyConvert(shape1));
         shapes.push_back(shapeToPolyConvert(shape2));

         break;
      }
      default:
      {
       /* Otherwise convert the single shape */
        shapes.push_back(shapeToPolyConvert(element));
      }
    }
    RETURN shapes;
}

/* Method: shapeToPolyConvert(FShape element)
   Purpose: converts a DSS shape into a polygon
*/
FShape * ConcreteArfDataServer::shapeToPolyConvert(FShape * element) const
{
  PROC;

   unsigned Maxpts = 20; // Maximum number of sides in the polygon
   unsigned Npts = 0;
   double angle = 0.0;
   double rotang = 0.0;  // The rotation angle (counter-clockwise !)
   double toang = 0.0;
   vector<float> x(Maxpts), y(Maxpts);

  /* Get centre of the shape */
    float cX = element->pointsX()[0];
    float cY = element->pointsY()[0];
    float trX = 0.0;
    float trY = 0.0;
    float a = 0.0;
    float b = 0.0;
    double degtorad = 3.14159265 / 180.0;
    float theta=0.0;
    float z=0.0;

  /* Get top left, top right, radii and rotation angles from the 
     input element. */ 

   switch(element->type())
     {
      case ShapeBasics::RECTANGLE :
      case ShapeBasics::ROTRECTANGLE :
       {

        // Get top right
         trX = element->pointsX()[1];
         trY = element->pointsY()[1];
         rotang = element->rotAngles()[0] * degtorad;

         a = trX - cX;
         b = trY - cY;

         break;
       }
      case ShapeBasics::BOX :
      case ShapeBasics::ROTBOX :
      case ShapeBasics::ELLIPSE :
      case ShapeBasics::RHOMBUS:
      case ShapeBasics::ROTRHOMBUS:
       {
         a = element->radii()[0];
         b = element->radii()[1];
         rotang = element->rotAngles()[0] * degtorad;

         break;
       }
      case ShapeBasics::SECTOR :
       {
         rotang = element->rotAngles()[0] * degtorad;
         toang = element->rotAngles()[1] * degtorad;

         break;
       }

      case ShapeBasics::CIRCLE :
       {
         a = element->radii()[0];
         b = a;

         break;
       }
      default:
              errstr<<"Source region shape cannot be processed"
	       << Sas::error("ShapeNotHandled");
     }

     errstr<<"Point 0: "<<cX<<" "<<cY
     <<"Radii: "<<a<<" "<<b
     <<"Angles "<<rotang<<" "<<toang
	 << Sas::message(Msg::App, Msg::Verbose);

   /* Go round the shape setting points */
   switch(element->type())
     {
      case ShapeBasics::BOX :
      case ShapeBasics::ROTBOX :
        {
       // The radii are actually diameters so divide by 2
           a = a / 2;
           b = b / 2;

           theta = atan(a/b);
           z = sqrt(b*b + a*a);

             Npts = 4;
             x[0] = cX - z * sin(theta-rotang);
             y[0] = cY - z * cos(theta-rotang);
             x[1] = cX - z * sin(theta+rotang);
             y[1] = cY + z * cos(theta+rotang);
             x[2] = cX + z * sin(theta-rotang);
             y[2] = cY + z * cos(theta-rotang);
             x[3] = cX + z * sin(theta+rotang);
             y[3] = cY - z * cos(theta+rotang);

             break;
        }
      case ShapeBasics::RHOMBUS:
      case ShapeBasics::ROTRHOMBUS:
        {
             theta=M_PI/2.0;

             Npts = 4;
             x[0] = cX - a * sin(theta-rotang);
             y[0] = cY - a * cos(theta-rotang);
             x[1] = cX - b * cos(theta-rotang);
             y[1] = cY + b * sin(theta-rotang);
             x[2] = cX + a * sin(theta-rotang);
             y[2] = cY + a * cos(theta-rotang);
             x[3] = cX + b * cos(theta-rotang);
             y[3] = cY - b * sin(theta-rotang);

             break;
        }
      case ShapeBasics::RECTANGLE :
      case ShapeBasics::ROTRECTANGLE :
        {
           theta = atan(a/b);
           z = sqrt(b*b + a*a);

             Npts = 4;
             x[0] = cX;
             y[0] = cY;
             x[1] = cX - b * sin(rotang);
             y[1] = cY + b * cos(rotang);
             x[2] = cX + z * sin(theta-rotang);
             y[2] = cY + z * cos(theta-rotang);
             x[3] = cX + a * cos(rotang);
             y[3] = cY + a * sin(rotang);

             break;
        }
      case ShapeBasics::ELLIPSE :
      case ShapeBasics::CIRCLE :
        {
             Npts = Maxpts;
             for (unsigned n=0; n<Npts; n++) {
               angle += 2.0 * M_PI / Npts;
               x[n] = cX + a * sin(angle);
               y[n] = cY + b * cos(angle);
             }
             break;
        }
      case ShapeBasics::SECTOR:
        {
             Npts = Maxpts;

         // Set the centre point
             x[0] = cX;
             y[0] = cY;

         // Set points along the arc
             a=30000.0;

             for (unsigned n=1; n<Npts; n++) {
               angle=rotang + (toang - rotang) * (n-1) / (Npts-1);
               x[n] = cX + a * sin(angle);
               y[n] = cY + a * cos(angle);

             }
             break;
        }
      case ShapeBasics::POLYGON :
        {
            x = element->pointsX(); 		
            y = element->pointsY(); 		
            Npts = x.size();

            break;
        }

      default:
              errstr<<"Wrong shape - report bug"
	       << Sas::error("ShapeNotHandled");
    }

    for (unsigned n=0; n<Npts; n++) {
         errstr<<"adding point: "<<x[n]<<" "<<y[n]<<" "<<theta
	 << Sas::message(Msg::App, Msg::Noisy);
    }
    
   /* Convert points to detector coordinates if neccessary */
    if (_regPosCoords) 
    {
       for (unsigned n=0; n<Npts; n++) {
           SourcePosition s = _coreData->xyToDetxy(x[n], y[n]);
           x[n] = s.xdet();
           y[n] = s.ydet();
               errstr<<"Det point: "<<x[n]<<" "<<y[n]
	       << Sas::message(Msg::App, Msg::Verbose);
       }
    }

   /* Return the converted shape and whether it is included/excluded */
    RETURN (new FPolygon(x, y, Npts, element->included()));
}

void ConcreteArfDataServer::createDetmap()
{
  /* Define variables to hold detmap bounds.
     The source position will be checked
     later to see if it is within these bounds
  */
  double srcxpos = _coreData->sourcePosition()->xdet();
  double srcypos = _coreData->sourcePosition()->ydet();

  /* Construct PixelIterator object */
  string detmaptype = stringParameter("detmaptype");
  unsigned xbins = (unsigned) intParameter("detxbins");
  unsigned ybins = (unsigned) intParameter("detybins");

  /* If source is extended and detmap is not a dataset, use a minimum of 
     10 bins */
  if ( detmaptype !="dataset" && (_extendedsource) )
    {
   //RDS - removing this bit of software facism
     //xbins = xbins < 10 ?10:xbins;
     //ybins = ybins < 10 ?10:ybins;

     errstr 
	<< "detector map binning will be: "
        << "xbins: "<<xbins <<" ybins: "<<ybins
	<< Sas::message(Msg::App, Msg::Sparse);
    }

  if ( detmaptype =="dataset" )
    {
      errstr 
	<< "Constructing a DATASET-based detector map."
	<< Sas::message(Msg::App, Msg::Sparse);

      
     /* Create ArrayPixelIterator object - add colon if necessary*/
      if (stringParameter("detmaparray").find(":") != string::npos) { 
         _rawDetmap = new ArrayPixelIterator(stringParameter("detmaparray"));
      } else {
         _rawDetmap = new ArrayPixelIterator(
                         stringParameter("detmaparray")+":");
      }
      
      /* Get WCS info for checks */
   //   const WCSConvertor * wcsconv = 
//	_rawDetmap->wcsconvertor();

      /* The WCS axes must be DETX and DETY resp. */
      /*
      if ( ( wcsconv->xType() != "DETX") || 
	   ( wcsconv->yType() != "DETY" ) )
	{
	  
	  errstr 
	    << "The detmap axes are "
	    << wcsconv->xType()
	    << " and "
	    << wcsconv->yType()
	    << ". These should be DETX and DETY"
	    << Sas::error("InvalidWCSAxes");
	}
      */
	  
	  
      /* Now need to access size */
      unsigned xsize = _rawDetmap->xbins();
      unsigned ysize = _rawDetmap->size()/xsize;
      
      
//        double xmin = wcsconv->physicalX(1);
//        double xmax = wcsconv->physicalX(xsize);
      
//        double ymin = wcsconv->physicalY(1);
//        double ymax = wcsconv->physicalY(ysize);
      
      /* Cast to PixelIterator type */
      //     _rawDetmap = arritr;

      CoordPair rawboundsmin = _rawDetmap->physXYmin();
      CoordPair rawboundsmax = _rawDetmap->physXYmax();

      errstr
        << "The dataset-based detmap bounds corresponds to "
        << " xmin="  << rawboundsmin.x()
        << " xmax="  << rawboundsmax.x()
        << " xbins=" << xsize
        << " ymin="  << rawboundsmin.y()
        << " ymax="  << rawboundsmax.y()
        << " ybins=" << ysize
        << Sas::message(Msg::App, Msg::Sparse);

      /* Check the whether the extent of the regions are larger
	 than that of the detector map */
  
      if(_extent->isValid())
	{
          _extent->checkBounds( rawboundsmin.x(),
                                rawboundsmax.x(),
                                rawboundsmin.y(),
                                rawboundsmax.y()
                                );
	}
    }
  else
    {
      /* PSF or FLAT detector map */

      double xmin = 0;
      double xmax = 0;
      double ymin = 0;
      double ymax = 0;

      if (booleanParameter("withdetbounds"))
	  {
	    errstr 
	      << "Taking detector map bounds from the parameters detxoffset and detyoffset ..."
	      << Sas::message(Msg::App, Msg::Sparse);

	    double xoffset = realParameter("detxoffset");
	    double yoffset = realParameter("detyoffset");

	    xmin = srcxpos-xoffset;
	    xmax = srcxpos+xoffset;
	    ymin = srcypos-yoffset;
	    ymax = srcypos+yoffset;
	  }
      else
	{
	  errstr 
	    << "Computing detector map bounds from the region selection information stored in the spectrum ..."
	    << Sas::message(Msg::App, Msg::Sparse);

	  if (_extent->isValid())
	    {
	      xmin = _extent->xmin();
	      xmax = _extent->xmax();
	      ymin = _extent->ymin();
	      ymax = _extent->ymax();
	    }
	  else
	    {
	      errstr 
		<< "Cannot determine bounds for detector map from region extent."
		<< "Either check DSS information in spectrum,"
		<< "or set the bounds explicitly by setting the parameter --withdetbounds to true," 
		<< " and entering appropriate values for --detxoffset and --detyoffset."
		<< Sas::error("RegionInvalid");
	    }
	}

      if (detmaptype=="psf")
	{
	    errstr 
	      << "Constructing a PSF detector map with "
	      << " xmin="  << xmin
	      << " xmax="  << xmax
	      << " xbins=" << xbins
	      << " ymin="  << ymin
	      << " ymax="  << ymax
	      << " ybins=" << ybins
	      << Sas::message(Msg::App, Msg::Sparse);


	  _rawDetmap = 
	    new PsfmodelPixelIterator( xmin,
				       xmax,
				       xbins,
				       ymin,
				       ymax,
				       ybins );
	  
	}
      else
	{
	  errstr 
	    << "Constructing a FLAT detector map with "
	    << " xmin="  << xmin
	    << " xmax="  << xmax
	    << " xbins=" << xbins
	    << " ymin="  << ymin
	    << " ymax="  << ymax
	    << " ybins=" << ybins
	    << Sas::message(Msg::App, Msg::Sparse);

	  _rawDetmap = 
	    new DummyPixelIterator( xmin,
				    xmax,
				    xbins,
				    ymin,
				    ymax,
				    ybins );
	}
    }
  

  /* Check that source position is within bounds of map */
  if (!_rawDetmap)
    {
      errstr<<"Anomaly in ConcreteArfDataServer - raise SPR"
          << Sas::error("ARFObjError");
    }

  CoordPair rawboundsmin = _rawDetmap->physXYmin();
  CoordPair rawboundsmax = _rawDetmap->physXYmax();
  double xmin = rawboundsmin.x();
  double xmax = rawboundsmax.x();
  double ymin = rawboundsmin.y();
  double ymax = rawboundsmax.y();

  if ( srcxpos < xmin || srcxpos > xmax ||
       srcypos < ymin || srcypos > ymax )
    {
      errstr 
	<< "The source position in DET coords (x="
	<< srcxpos
	<< " y="
	<< srcypos
	<< ") is outside the bounds of the detector map (xmin="
	<< xmin
	<< " xmax="
	<< xmax
	<< " ymin="
	<< ymin
	<< " ymax="
	<< ymax
	<< "). Please specify a valid source position or detector map."
	<< Sas::error("InvalidSourcePosition");
    }

  ENDPROC;
}
/* addDetRegion - adds a pseudo spatial region in detector coords to the dss */
void ConcreteArfDataServer::addDetRegion(PositionUtils * putils)
{
  PROC;

  errstr<<"Entered addDet"
    << Sas::message(Msg::App, Msg::Verbose);

  // Keep this code as it may prove useful but set a very simple
  //   detector coordinate region //

  // Find the extents of the DSS in RAW-X 
  double lower_rawx=0, lower_rawy=0;
  double upper_rawx=0, upper_rawy=0;
  RangeBound lowerB_rawx;
  RangeBound upperB_rawx;

  // DSS::DSaxes axestest;
  // axestest.push_back("RAW");

  Range<double> extent_rawx = _dss->extent("RAWX");
  extent_rawx.values(lower_rawx,
                     upper_rawx,
                     lowerB_rawx,
                     upperB_rawx);

  errstr
    << "RAWX extents : "
    << " lower_rawx=" << lower_rawx
    << " lowerB_rawx=" << (unsigned) lowerB_rawx
    << " upper_rawx=" << upper_rawx
    << " upperB_rawx=" << (unsigned) upperB_rawx
    << Sas::message(Msg::App, Msg::Verbose);

 // If the raw-x range is unbounded set reasonable defaults 
  if ( lowerB_rawx == Unbounded || upperB_rawx == Unbounded ) {

     switch (_windowMode) {

      // PN - timing mode
        case Xmm::FAST_TIMING:      
          {
            lower_rawx = 1;
            upper_rawx = 64;
            break;
          }
      // PN - burst mode
        case Xmm::FAST_BURST:
          {
            lower_rawx = 1;
            upper_rawx = 64;
            break;
          }
      // MOS - timing mode
        case Xmm::FAST_UNCOMPRESSED:
          {
            lower_rawx = 250;
            upper_rawx = 350;
            break;
          }
      // MOS - burst mode
        case Xmm::FAST_COMPRESSED:
          {
            lower_rawx = 250;
            upper_rawx = 350;
            break;
          }
        default:
          errstr<<"Some problem with timing/burst modes - raise SPR"
              << Sas::message(Msg::App, Msg::Sparse);
        }
    }

   // Set the range for raw-y depending on the instrument 
    switch (_instrumentID)
    {
    case Xmm::EMOS1 :
    case Xmm::EMOS2 :
      {
        lower_rawy = 1;
        upper_rawy = 600;
        break;
      }
    case Xmm::EPN :
      {
        lower_rawy = 1;
        upper_rawy = 200;
        break;
      }
    default:
        errstr<<"Instrument "<<_instrumentID<<" not recognised"
              << Sas::error("invalidInstrument");
    }

   // Now find the detector coords of this position
    PixCoords pixCoord1(lower_rawx, lower_rawy);
    CamCoords camCoord2 = putils->PixCoordTocamCoord2(pixCoord1);
    double det_xlow = putils->CamCoord2Todetxy005(camCoord2.x());
    double det_ylow = putils->CamCoord2Todetxy005(camCoord2.y());
    PixCoords pixCoord2(upper_rawx, upper_rawy);
    camCoord2 = putils->PixCoordTocamCoord2(pixCoord2);
    double det_xupp = putils->CamCoord2Todetxy005(camCoord2.x());
    double det_yupp = putils->CamCoord2Todetxy005(camCoord2.y());

    double detxcen = (det_xlow + det_xupp) / 2.0;
    double detycen = (det_ylow + det_yupp) / 2.0;
    double detx_hwid = abs(detxcen - det_xlow);
    double dety_hwid = abs(detycen - det_ylow);

   // Create an expression from these coords 
    std::string expression;
    expression="(DETX,DETY) in BOX("+toString(detxcen)+","+toString(detycen)+
               ","+toString(detx_hwid)+","+toString(dety_hwid);
    expression += ",0)";

    errstr<<"From raw X: "<<lower_rawx<<" "<<upper_rawx
              <<" raw Y: "<<lower_rawy<<" "<<upper_rawy<<" got: "
              <<expression
              << Sas::message(Msg::App, Msg::Verbose);

   /* In general the spatial information is lost in timing mode;
      just dummy up a simple, near to the boresight region and
      ignore the preceeding code */

    expression="(DETX,DETY) in BOX(100,100,100,100,0)";

   /* Create a filter */
    DataSubspace _dssextra = dssBuilder.getDss(expression);
   
    for(unsigned icomp=0; icomp <_dssextra.components(); icomp++){

      errstr<<"Adding component "<<icomp<<" to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

      _dss->addComponent(*_dssextra.component(icomp));
    }

    errstr<<"Added new component to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

 ENDPROC;
}

/* Method readDSS() - reads some values from the data subspace */
void ConcreteArfDataServer::readDSS()
{
  PROC;

    DSS::DSaxes axestest;
    axestest.push_back("FLAG");

     _flag=0xffffffff;

   /* Read the FLAG value if any from the DSS */
     for (unsigned i  = 0; i < _dss->components(); ++i) {

       DScomp * dscomp = _dss->component(i);

     /* Loop over each filter in this component */
       for (unsigned f = 0; f < dscomp->filters(); ++f) {

          DSfilter * dsfilt = dscomp->filter(f);

        /* Try and find a component which has a FLAG filter */
          if (dsfilt->name() == "FLAG") 
          {

           /* Have to handle masks and ranges separately */
            if (dsfilt->type() == DataSubspace::BitMaskT) {
               _flag &= dsfilt->bitMask();
            }
            else if (dsfilt->type() == DataSubspace::RangeT) {

               list <Range<double> > ranges = dsfilt->ranges();

               double upper, lower;
               RangeBound upperB, lowerB;

             /* Treat this mask as a range, even though its a single value */
               for ( list<Range<double> >::iterator i = ranges.begin();
                  i != ranges.end(); ++i )
               {
                  (*i).values(lower, upper, lowerB, upperB);
               }
               _flag &= (unsigned long) lower;
            }
            errstr<<"Setting BIT mask "<<_flag
                << Sas::message(Msg::App, Msg::Verbose);

         }
       } 

     }

  ENDPROC;
}



