//
// Module: RectanglePsfModel
// Author: Richard Saxton
//
// Description: 
//
// A simple model for the PSF.
// Used for prototyping; will be eventually replaced by relevant CAL fn.
//
// Modifications: See ChangeLog
// 
// $Id: RectanglePsfModel.h,v 1.1 2000/10/19 15:12:49 hsiddiqu Exp $

#ifndef RectanglePsfModel_H
#define RectanglePsfModel_H

/* SAS-specific */
#include "CalCommon.h"
#include "CalServer.h"
#include "EncircledEnergyDataServer.h"

/* Local */
#include "CirclePsfModel.h"

class RectanglePsfModel: public CirclePsfModel
{
public:

  /*** NOTE: units for both x/y directions are 0.05 arcseconds ***/

  /* Constructor */
   RectanglePsfModel(  double Energy,             /* Psf energy */
                    double Theta,              /* Off-axis angle */
                    double Phi,                /* Azimuthal angle */
                    vector<double> HalfWidth,  /* Box half-widths (asec) */
                    unsigned int NGridEls      /* 100 for 10*10 det map. */
                    );    

  /* Destructor */
  ~RectanglePsfModel();

  /* Calculate integrated PSF over a square area */
  double int_psf( double xboxcentre, 
		  double yboxcentre, 
		  double xboxsize,
		  double yboxsize) const;

  /* Calculate the PSF integrated within a rectangle */
  double rad_psf() const;

  string mode() const;

private:
  
  double _energy;
  double _theta;

  double _phi;
  vector<double> _radii;
  double _diagonal;
  unsigned int _ngridels;

  EncircledEnergyDataServer * _eds;

  EncircledEnergyFunction *eef;
};

#endif

